module Geometry (Vec(..), Point, rand_norm_vec, (%+), (%-), Ray, (+*), LTr(..), tr, ATr(..)) where

import Algebra
import Applicable
import Rand

infixl 6 %+, %-, +*

data Vec = V {-# UNPACK #-} !Double {-# UNPACK #-} !Double {-# UNPACK #-} !Double deriving Show

instance Vector Vec where
	zvec = V 0 0 0
	(V x0 x1 x2) @+ (V y0 y1 y2) = V (x0 + y0) (x1 + y1) (x2 + y2)
	(V x0 x1 x2) .* a = V (x0 * a) (x1 * a) (x2 * a)

instance IPVector Vec where
	(V x0 x1 x2) @. (V y0 y1 y2) = (x0 * y0) + (x1 * y1) + (x2 * y2)

instance NormVector Vec where
	normSqr (V x0 x1 x2) = x0 ^ 2 + x1 ^ 2 + x2 ^ 2

rand_norm_vec :: (RandomGenD g) => Rand g Vec
rand_norm_vec =
	do	x <- rand01
		y <- rand01
		z <- rand01
		let v = V (x - 0.5) (y - 0.5) (z - 0.5) in
			if normSqr v > 0.25 then
				rand_norm_vec
			else
				return (normalize v)

type Point = Vec

(%+) :: Point -> Vec -> Point
(%+) = (@+)

(%-) :: Point -> Point -> Vec
(%-) = (@-)

type Ray = (Point, Vec)

(+*) :: Ray -> Double -> Point
(base, dir) +* dist = base %+ dir .* dist

-- each Vec is a column of the matrix
data LTr = L !Vec !Vec !Vec deriving Show

instance Vector LTr where
	zvec = L zvec zvec zvec
	(L a0 a1 a2) @+ (L b0 b1 b2) = (L (a0 @+ b0) (a1 @+ b1) (a2 @+ b2))
	(L a0 a1 a2) .* b = (L (a0 .* b) (a1 .* b) (a2 .* b))

instance Algebra LTr where
	a @* (L b0 b1 b2) =
		L (apply a b0) (apply a b1) (apply a b2)

instance AlgebraU LTr where
	idvec = (L (V 1 0 0) (V 0 1 0) (V 0 0 1))

instance Applicable LTr Vec Vec where
	apply (L a0 a1 a2) (V v0 v1 v2) = a0 .* v0 @+ a1 .* v1 @+ a2 .* v2

instance Invertible LTr where
	inv t@(L (V t00 t01 t02) (V t10 t11 t12) (V t20 t21 t22)) = t' ./ det t
		where
			t' = L
				(V (t11 * t22 - t12 * t21) (t12 * t20 - t10 * t22) (t10 * t21 - t11 * t20))
				(V (t21 * t02 - t22 * t01) (t22 * t00 - t20 * t02) (t20 * t01 - t21 * t00))
				(V (t01 * t12 - t02 * t11) (t02 * t10 - t00 * t12) (t00 * t11 - t01 * t10))

tr :: LTr -> LTr
tr (L (V t00 t01 t02) (V t10 t11 t12) (V t20 t21 t22)) =
	L (V t00 t10 t20) (V t01 t11 t21) (V t02 t12 t22)

det :: LTr -> Double
det (L (V t00 t01 t02) (V t10 t11 t12) (V t20 t21 t22)) =
		t00 * t11 * t22
	-	t00 * t12 * t21
	-	t01 * t10 * t22
	+	t01 * t12 * t20
	+	t02 * t10 * t21
	-	t02 * t11 * t20

data ATr = A {linear :: !LTr, translation :: !Vec} deriving Show

instance Vector ATr where
	zvec = A zvec zvec
	(A l t) @+ (A l' t') = A (l @+ l') (t @+ t')
	(A l t) .* a = A (l .* a) (t .* a)

instance Algebra ATr where
	(A l t) @* (A l' t') = A (l @* l') (apply l t' @+ t)

instance AlgebraU ATr where
	idvec = A idvec zvec

instance Applicable ATr Vec Vec where
	apply (A l t) x = apply l x @+ t

instance Applicable ATr Ray Ray where
	apply a (base, dir) = (apply a base, apply (linear a) dir)

instance Invertible ATr where
	inv (A l t) = A (inv l) (apply (inv l) t .* (-1))
