module Material (Material(..), SurfaceMat, hollow_mat, color_surface, ibdrf_to_surf_mat, lambert, ideal_specular, fuzzy_specular, fresnel, shape_to_part, checker, add_surf_mat) where

import Algebra
import Color
import Geometry
import qualified LightProp as LP
import Math
import Maybe
import Monad
import Part
import Rand
import Shape
import Types

data (RandomGenD g) => Material g =
		Mat
		{	surface_mat :: SurfaceMat g
		,	inside_dim, outside_dim :: Ray -> Double -> LP.T g
		,	ior :: Double
		}

-- Input: IOR ratio (inside / outside), point, normal
-- output: Nothing if surface is completely transparent; otherwise Just an EBDRF
type SurfaceMat g = Double -> Point -> Vec -> Maybe (EBDRF g)

type IBDRF g = Double -> Vec -> Vec -> RandMeas g Vec Double

hollow_mat :: (RandomGenD g) => SurfaceMat g -> Material g
hollow_mat surface_mat = Mat surface_mat idvec idvec 0

color_surface :: (RandomGenD g) => Color -> SurfaceMat g -> SurfaceMat g
color_surface color surface_mat ior loc normal = liftM (flip mult_rm (LP.C color Nothing) .) (surface_mat ior loc normal)

reflect :: Vec -> Vec -> Vec
reflect normal dir = dir @- normal .* (dir @. normal * 2)

lambert :: (RandomGenD g) => IBDRF g
lambert = const lambert_bdrf

lambert_bdrf :: (RandomGenD g) => Vec -> Vec -> RandMeas g Vec Double
lambert_bdrf normal in_dir = ac_randmeas dir_f dir_r where
	dir_r =
		do	(cos_r, dir) <- rand_cond ((0 <) . norm . fst) dir_r'
			return (2 * cos_r, dir)
	dir_r' =
		do
			dir <- rand_norm_vec
			return (normal' @. dir, dir)
	dir_f = max 0 . (4 *) . (normal' @.)
	normal' = if 0 > normal @. in_dir then normal else normal .* (-1)

ideal_specular :: (RandomGenD g) => IBDRF g
ideal_specular = const ideal_specular_bdrf

ideal_specular_bdrf :: (RandomGenD g) => Vec -> Vec -> RandMeas g Vec Double
ideal_specular_bdrf normal in_dir = m where
	m = ms_randmeas $ return (1, out_dir)
	out_dir = reflect normal in_dir

-- TODO: there should be a normalizing constant here!
fuzzy_specular :: (RandomGenD g) => Double -> IBDRF g
fuzzy_specular shininess _ normal in_dir = m where
	m = ac_randmeas ac_f (liftM (\x -> (ac_f x, x)) rand_norm_vec)
	ac_f x = if normal @. x > 0 && out_dir @. x > 0 then (out_dir @. x) ** shininess else 0
	out_dir = reflect normal in_dir

-- from Wikipedia's article ``Fresnel equations''
fresnel :: (RandomGenD g) => IBDRF g
fresnel ior_ratio normal in_dir = m where
	m = ms_randmeas out_dir

	out_dir =
		do	dir <- fuzzy_if reflection
				(return (reflect normal in_dir))
				(return transmit_dir)
			return (1, dir)

	reflection = (refl_s + refl_p) / 2

	-- reflection coefficient for light polarized
	-- perpendicular to the plane of refraction
	refl_s =
		((cos_in - ior_ratio' * cos_out) /
		(cos_in + ior_ratio' * cos_out)) ^ 2

	-- reflection coefficient for light polarized
	-- parallel to the plane of refaction
	refl_p =
		((cos_out - ior_ratio' * cos_in) /
		(cos_out + ior_ratio' * cos_out)) ^ 2

	-- this Should end up having norm 1
	transmit_dir = (in_dir @- normal' .* (cos_out * ior_ratio' - cos_in)) ./ ior_ratio'

	cos_in = - normal' @. in_dir
	sin_in = sqrt (1 - cos_in ^ 2)
	sin_out = sin_in / ior_ratio
	cos_out = sqrt (1 - sin_out ^ 2)

	(normal', ior_ratio') =
		if 0 <= in_dir @. normal
		then (normal .* (-1), 1 / ior_ratio)
		else (normal, ior_ratio)

ibdrf_to_surf_mat :: (RandomGenD g) => (Double -> Vec -> Vec -> RandMeas g Vec Double) -> SurfaceMat g
ibdrf_to_surf_mat ior_to_bdrf ior_ratio _ normal = Just (either (const zero_emit) m') where
	m in_dir = ior_to_bdrf ior_ratio normal in_dir
	m' in_dir =
		RM
			(ac_p_ (m in_dir))
			(either (const zvec) ((idvec .*) . (ac_f_ (m in_dir))))
			(liftM (\(v, d) -> (idvec .* v, Right d)) (ac_r_ (m in_dir)))
			(either (error "iorbdrf_to_surface_mat: measure undefined on Left") (ac_rf_ (m in_dir)))
			(liftM (\(v, d) -> (idvec .* v, Right d)) (ms_r_ (m in_dir)))

shape_to_part :: (RandomGenD g) => Shape -> Material g -> Part g
shape_to_part shape mat = Part trace where
	-- TODO: sample from the surface of the shape if it emits light
	trace (Left _) = (MaybeT $ return Nothing, error "shape_to_part: dim from source/sink undefined")

	trace (Right (base, in_dir)) = (hit, dim) where
		intersections = Shape.trace (surface_ shape) (base, in_dir)
		first_intersection = listToMaybe intersections
		hit = MaybeT $ return $
			do	(dist, int@(base, normal)) <- first_intersection
				ebdrf <- surface_mat mat (ior mat) base normal
				return $ Hit dist base ebdrf
		dim dist = vprod [which_dim (point, in_dir) dist | (point, dist, which_dim) <- zip3 points dists which_dims] where
			interesting_ints = takeWhile ((dist - eps >) . fst) intersections
			points = base : map (fst . snd) interesting_ints
			dists = pair_dists (0 : map fst interesting_ints ++ [dist])
			pair_dists (a : t@(b : _)) = b - a : pair_dists t
			pair_dists _ = []
			which_dims = cycle $ (if inside_ shape base then id else reverse) $ map ($ mat) [inside_dim, outside_dim]

checker :: (RandomGenD g) => SurfaceMat g -> SurfaceMat g -> SurfaceMat g
checker black white ior loc@(V x y z) = (if mod (floor x + floor y + floor z) 2 == 0 then black else white) ior loc

add_surf_mat :: (RandomGenD g) => SurfaceMat g -> SurfaceMat g -> SurfaceMat g
add_surf_mat x y ior loc normal = Just (fromJust (x ior loc normal) @+ fromJust (y ior loc normal))
