module Rand (RandomGenD(..), Rand(..), RandMeas(..), stdgen_is_good, fuzzy_if, rm_sample, rm_sample_tell, unif_randmeas, ac_randmeas, ms_randmeas, mult_rm, point_mass, rand01, rand_n, rand_norm, rand_exp, rand_cond, rand_list, bias_bool, bypass) where

import Algebra
import Monad
import Random

class RandomGenD g where
-- nextd should return uniformly distributed random numbers from the interval [0, 1)
	nextd :: g -> (Double, g)
	splitd :: g -> (g, g)

instance RandomGenD StdGen where
	nextd g = (fromIntegral i / stdgen_range, g')
		where
			(rmin, rmax) = genRange g
			(i, g') = next g
	splitd = split

stdgen_range :: Double
stdgen_range = fromIntegral rmax + 1
	where
		(_, rmax) = genRange (undefined :: StdGen)

stdgen_is_good :: Bool
stdgen_is_good = range_is_okay && (rmin == 0)
	where
		(rmin, _) = genRange (undefined :: StdGen)
		range_is_okay = stdgen_range > 2^31 * 0.9 && stdgen_range * 0.9 < 2^31

newtype (RandomGenD g) => Rand g a = R { rfun :: g -> (a, g) }

{-
	A RandMeas g d v is a measure over d taking values in the vector space v.  An existing measure is assumed to be defined over d.  The measure is the sum (ac_p_ * M + (1 - ac_p_) * N), where M is absolutely continuous and N is mutually singular to the existing measure on v.  M is ac_f_ times the measure on v.  M can be sampled with ac_r_ and N can be sampled with ms_r_.  ac_rf_ gives the ratio of the value returned by ac_r_ divided by the value returned by ac_f_: proportional to ac_r_ v-values and inversely proportional to ac_f_.
-}
data (RandomGenD g, Vector v) => RandMeas g d v =
		RM
		{	ac_p_ :: Double
		,	ac_f_ :: d -> v
		,	ac_r_ :: Rand g (v, d)
		,	ac_rf_ :: d -> Double
		,	ms_r_ :: Rand g (v, d)
		}

instance (RandomGenD g, Vector v) => Vector (RandMeas g a v) where
	zvec = RM 1 (const zvec) zerr (error "ac_fr undefined for the zero measure") zerr
		where
			zerr = error "can't sample from the zero measure"
	u @+ v = RM ac_p ac_f ac_r ac_rf ms_r
		where
			ac_p = (ac_p_ u + ac_p_ v) / 2
			ac_f = (ac_f_ u @+ ac_f_ v) .* 2
			ac_r = fuzzy_if (ac_p_ u / (ac_p_ u + ac_p_ v)) (ac_r_ u) (ac_r_ v)
			ac_rf = (ac_rf_ u .* ac_p_ u @+ ac_rf_ v .* ac_p_ v) ./ (ac_p_ u + ac_p_ v)
			ms_r = fuzzy_if ((1 - ac_p_ u) / (2 - ac_p_ u - ac_p_ v)) (ms_r_ u) (ms_r_ v)
	(RM ac_p ac_f ac_r ac_rf ms_r) .* c = RM ac_p (ac_f .* c) (f ac_r) ac_rf (f ms_r)
		where
			f r = do
				(magnitude, point) <- r
				return (magnitude .* c, point)

fuzzy_if :: (RandomGenD g) => Double -> Rand g a -> Rand g a -> Rand g a
fuzzy_if p if_yes if_no = do
	yes <- bias_bool p
	if yes then if_yes else if_no

rm_sample :: (RandomGenD g, Vector v) => RandMeas g d v -> Rand g (v, d)
rm_sample m =
	do	ac <- bias_bool (ac_p_ m)
		(if ac then ac_r_ else ms_r_) m

rm_sample_tell :: (RandomGenD g, Vector v) => RandMeas g d v -> Rand g (Bool, v, d)
rm_sample_tell m =
	do	ac <- bias_bool (ac_p_ m)
		(i, x) <- (if ac then ac_r_ else ms_r_) m
		return (ac, i, x)

unif_randmeas :: (RandomGenD g, AlgebraU v) => Rand g d -> RandMeas g d v
unif_randmeas = ac_randmeas idvec . liftM (\x -> (idvec, x))

ac_randmeas :: (RandomGenD g, Vector v) => (d -> v) -> Rand g (v, d) -> RandMeas g d v
ac_randmeas f r = RM 1 f r idvec (error "ac_randmeas: ms_r undefined")

ms_randmeas :: (RandomGenD g, Vector v) => Rand g (v, d) -> RandMeas g d v
ms_randmeas r =
	RM
		0
		zvec
		(error "ms_randmeas: ac_r undefined")
		(error "ms_randmeas: ac_rf undefined")
		r

mult_rm :: (RandomGenD g, Algebra v) => RandMeas g d v -> v -> RandMeas g d v
mult_rm (RM ac_p ac_f ac_r ac_rf ms_r) x =
	RM
		ac_p
		((x @*) . ac_f)
		(liftM (\(y, d) -> (x @* y, d)) ac_r)
		ac_rf
		(liftM (\(y, d) -> (x @* y, d)) ms_r)

point_mass :: (RandomGenD g, Vector v) => d -> v -> RandMeas g d v
point_mass p m = ms_randmeas (return (m, p))

instance (RandomGenD g) => Monad (Rand g) where
	(R r) >>= f = R (\g -> let {(a, g') = r g ; R s = f a} in s g')
	return a = R (\g -> (a, g))

rand01 :: (RandomGenD g) => Rand g Double
rand01 = R nextd

rand_n :: (RandomGenD g) => Int -> Rand g Int
rand_n n = liftM (floor . (fromIntegral n *)) rand01

rand_norm :: (RandomGenD g) => Rand g Double
-- from http://emme2.spiess.ch/e2news/news14/node4.html 20070423
rand_norm = do
	u1 <- rand01
	u2 <- rand01
	return (sqrt (-2 * log u1) * cos (2 * pi * u2))

-- exponential distribution with mean 1
rand_exp :: (RandomGenD g) => Rand g Double
rand_exp = do
	x <- rand01
	return (- log x)

rand_cond :: (RandomGenD g) => (a -> Bool) -> Rand g a -> Rand g a
rand_cond f r = do
	x <- r
	if f x then return x else rand_cond f r

rand_list :: (RandomGenD g) => Rand g a -> Rand g [a]
rand_list r = bypass l
	where
		l = do
			h <- r
			t <- l
			return (h : t)

charf :: (Eq a) => a -> a -> Double
charf x y = if x == y then 1 else 0

bias_bool :: (RandomGenD g) => Double -> Rand g Bool
bias_bool p =
	do	x <- rand01
		return (x < p)

bypass :: (RandomGenD g) => Rand g a -> Rand g a
bypass r =
	do	g <- R splitd
		return $ fst $ rfun r g
