module Render (render, RenderCommand(..), RenderResponse(..), width, height) where

import Algebra
import Color

import Control.Concurrent

import Data.Array
import Data.Array.Storable
import Data.IORef
import Data.Monoid

import Geometry
import GlobalInfo
import IO
import Image
import qualified LightProp as LP
import Maybe
import Metropolis
import Rand
import Random
import Types
import Util

data RenderCommand = SendUpdate {rc_clear :: Bool}

data RenderResponse =
		RR
		{	rr_image :: Array (Int, Int) LongRGB
		,	rr_last_path :: [Point]
		,	rr_last_color :: RGB
		,	rr_n_mutations :: Integer
		,	rr_longest_path :: Int
		,	rr_n_mutations_accepted :: Integer
		}

data RenderState g =
		RS
		{	command_chan :: !(Chan RenderCommand)
		,	response_chan :: !(Chan RenderResponse)
		,	image :: !(StorableArray (Int, Int) LongRGB)
		,	points :: ![(RGB, (LP.T g, [Point]), Bool)]
		,	n_mutations :: !Integer
		,	longest_path :: !Int
		,	n_mutations_accepted :: !Integer
		}

width :: (Num a) => a
width = 300

height :: (Num a) => a
height = 300

points_per_step :: (Num a) => a
points_per_step = 2^11

mid3 :: (a, b, c) -> b
mid3 (_, x, _) = x

render_step :: RenderState g -> IO (RenderState g)
render_step state =
	let (batch, rest) = splitAt (2 * points_per_step) (points state) in do
		sequence_ [nplot (image state) (rgbToLongRGB color) x y | (color, (lp, _), _) <- batch, Just (x, y) <- [LP.pixel lp]]
		state' <- return $
			state
			{	points = rest
			,	n_mutations = points_per_step + n_mutations state
			,	longest_path = max
					(longest_path state)
					(maximum (map (length . snd . mid3) batch))
			,	n_mutations_accepted = n_mutations_accepted state + sum [indic new | (_, _, new) <- batch]
			}
		respond state' (head batch)

rgbs :: [Double] -> [RGB]
rgbs [] = []
rgbs (r : g : b : t) = RGB r g b : rgbs t

respond :: RenderState g -> (RGB, (LP.T g, [Point]), Bool) -> IO (RenderState g)
respond state (_, (lp, path), _) =
	do	empty <- isEmptyChan (command_chan state)
		if empty
			then return state
			else
				do	i_image <- freeze (image state) :: IO (Array (Int, Int) LongRGB)
					respond' state
						( RR
							i_image
							path
							(LP.color lp)
							(n_mutations state)
							(longest_path state)
							(n_mutations_accepted state)
						)

respond' :: RenderState g -> RenderResponse -> IO (RenderState g)
respond' state response =
	do	SendUpdate {rc_clear} <- readChan (command_chan state)
		writeChan (response_chan state) response
                state' <-
                    if rc_clear
                        then do fresh_image <- Data.Array.Storable.newArray ((0, 0), (width - 1, height - 1)) mempty
                                return $
                                    state
                                    { image = fresh_image
                                    , n_mutations = 0
                                    , longest_path = 0
                                    , n_mutations_accepted = 0
                                    }
                        else return state
		empty <- isEmptyChan (command_chan state')
		if empty
			then return state'
			else respond' state' response

init :: GlobalInfo -> MutationScheme StdGen d RGB (LP.T g, [Point]) -> Chan RenderCommand -> Chan RenderResponse -> IO (RenderState g)
init global muts command_chan response_chan = do
	image <- Data.Array.Storable.newArray ((0, 0), (width - 1, height - 1)) mempty
	path_ref <- newIORef []
	gen <- newStdGen -- TODO: let user control random seed
	return (RS command_chan response_chan image (fst (rfun (metropolis_samples muts) gen)) 0 0 0)

ouroboros :: (a -> IO a) -> a -> IO b
ouroboros f x = f x >>= ouroboros f

render :: MutationScheme StdGen d RGB (LP.T g, [Point]) -> GlobalInfo -> Chan RenderCommand -> Chan RenderResponse -> IO ()
render muts global command_chan response_chan =
	do	initial_state <- Render.init global muts command_chan response_chan
		ouroboros render_step initial_state
