module Shape (Surface(..), Shape(..), box, intersection, intersection_prep, plane, sphere, cylinder, cone) where

import Algebra
import Applicable
import Geometry
import Math
import Maybe
import Rand
import Util

-- trace returns a list of intersections, in increasing order of distance.
-- Each element is a pair (distance, normal).
-- If this is the surface of a Shape, the intersections returned by intersect_
-- must alternate between entering and leaving the Shape.  (If the intersection
-- just kissed the shape but the ray remains on the same side, exclude it
-- from the returned list.)
data Surface = Surf {trace :: Ray -> [(Double, Ray)]}

data Shape = Shape {surface_ :: Surface, inside_ :: Point -> Bool}

prepify :: Shape -> Prep ATr Shape
prepify (Shape (Surf hitf) in_f) = p idvec
	where
		p (inv_t :: ATr) = Prep a s' :: Prep ATr Shape
			where
				a t' = p (inv_t @* inv t')
				s' = Shape (Surf hitf') in_f'
				hitf' ray = map (trans_hit ray) (hitf (apply inv_t ray))
				trans_hit ray (t, (_, normal)) = (t, (ray +* t, normalize (apply (tr (linear inv_t)) normal)))
				in_f' = in_f . apply inv_t

box :: Prep ATr Shape
box = intersection_prep [plane xb 1, plane (xb .* (- 1)) 1, plane yb 1, plane (yb .* (- 1)) 1, plane zb 1, plane (zb .* (- 1)) 1]
	where
		xb = V 1 0 0
		yb = V 0 1 0
		zb = V 0 0 1

intersection :: [Shape] -> Shape
intersection shapes = Shape (Surf hitf) in_f
	where
		hitf ray@(_, dir) =
			let	all_hits = merge_many_by (\ (a, _) (b, _) -> compare a b) (map (flip trace ray . surface_) shapes)
			in	filter (\ (_, (p, _)) -> in_f (p @+ (dir .* eps))) all_hits
		in_f p = all (flip inside_ p) shapes

intersection_prep :: [Prep ATr Shape] -> Prep ATr Shape
intersection_prep shapes = Prep a i
	where
		a t = intersection_prep (map (flip papply t) shapes)
		i = intersection (map pget shapes)

plane :: Vec -> Double -> Prep ATr Shape
plane normal dist = Prep a surf
	where
		a (A l t) = plane normal' (dist * norm tnormal + normal' @. t)
			where
				tnormal = apply (inv (tr l)) normal
				normal' = normalize tnormal
		surf = Shape (Surf hitf) in_f
		hitf ray@(base, dir) = if hit_dist > eps then [hit] else []
			where
				hit_dist = (dist - normal @. base) / normal @. dir
				hit = (hit_dist, (ray +* hit_dist, normal))
		in_f p = p @. normal < dist

sphere :: Prep ATr Shape
sphere = prepify sphere'

-- unit sphere at origin
sphere' :: Shape
sphere' = Shape (Surf hitf) in_f
	where
		hitf ray@(base, dir) = l
			where
				a = normSqr dir
				b = 2 * (base @. dir)
				c = normSqr base - 1
				d = b ^ 2 - 4 * a * c
				l = if d > eps
					then [(t, (error "sphere: ray base", ray +* t)) | t <- [t0, t1], t > eps]
					else []
				t0 = (- b - r) / (2 * a)
				t1 = (- b + r) / (2 * a)
				r = sqrt d
		in_f p = normSqr p < 1

cylinder :: Prep ATr Shape
cylinder = prepify cylinder'

-- radius 1 cylinder along z axis
cylinder' :: Shape
cylinder' = Shape (Surf hitf) in_f
	where
		hitf ray@(base@(V bx by _), dir@(V dx dy _)) = l
			where
				a = dx ^ 2 + dy ^ 2
				b = 2 * (bx * dx + by * dy)
				c = bx ^ 2 + by ^ 2 - 1
				d = b ^ 2 - 4 * a * c
				l = if d > eps
					then [(t, (error "cylinder: ray base", n t)) | t <- [t0, t1], t > eps]
					else []
				t0 = (- b - r) / (2 * a)
				t1 = (- b + r) / (2 * a)
				r = sqrt d
				n t = let V x y _ = ray +* t in V x y 0
		in_f (V x y _) = x ^ 2 + y ^ 2 < 1

cone :: Prep ATr Shape
cone = prepify cone'

-- right cone along z axis
-- normal not normalized
cone' :: Shape
cone' = Shape (Surf hitf) in_f
	where
		hitf ray@(base@(V bx by bz), dir@(V dx dy dz)) = l
			where
				a = dx ^ 2 + dy ^ 2 - dz ^ 2
				b = 2 * (bx * dx + by * dy - bz * dz)
				c = bx ^ 2 + by ^ 2 - bz ^ 2
				d = b ^ 2 - 4 * a * c
				l = if d > eps
					then [(t, (error "cone: ray base", n t)) | t <- [t0, t1], t > eps]
					else []
				t0 = (- b - r) / (2 * a)
				t1 = (- b + r) / (2 * a)
				r = sqrt d
				n t = let V x y z = ray +* t in V x y (-z)
		in_f (V x y z) = x ^ 2 + y ^ 2 < z ^ 2
