module Shell (shell) where

import Char
import Color
import Control.Concurrent
import Data.Array.Unboxed
import Data.IORef

import Foreign.C.Types
import Foreign.Marshal.Array
import Foreign.Ptr

import Geometry
import GlobalInfo
import IO
import Image
import qualified LightProp as LP
import Math
import Metropolis
import Monad
import Part
import RList
import Rand
import Random
import qualified RListSample as RLS
import Render
import qualified SubpathSample as SPS
import System.CPUTime
import TheScenes
import Time

data RenderInfo =
		RI
		{	command_chan :: Chan RenderCommand
		,	response_chan :: Chan RenderResponse
		,	last_path_ref :: IORef [Point]
		,	image_file :: Handle
		,	start_cpu_time :: Integer -- in picoseconds
		,	start_clock_time :: Double -- in seconds
		-- GLUT ,	image_data :: Ptr CUChar
		-- GLUT ,	window :: Window
		}

data ShellState =
		SS
		{	global_info :: GlobalInfo
		,	render_info :: Maybe RenderInfo
		,	exit :: Bool
		}

data Lexeme = LParen | RParen | Semicolon | Ident String deriving Show

data BaseVal =
		VIint Int
	|	VS String
	|	VScene (forall g . (RandomGenD g) => Scene g)
	|	VIO (ShellState -> IO ShellState)
	|	VError String

data Val =
		Val
		{	val_f :: Val -> Val
		,	val_base :: BaseVal
		}

error_val :: String -> Val
error_val s =  Val (const (error_val s)) (VError s)

base_val :: BaseVal -> Val
base_val v = Val (const (error_val "base value applied as a function")) v

is_ident_char :: Char -> Bool
is_ident_char c = c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z' || c >= '0' && c <= '9' || c == '_' || c == '\''

lex :: String -> [Lexeme]
lex [] = []
lex ('(' : t) = LParen : Shell.lex t
lex (')' : t) = RParen : Shell.lex t
lex (';' : t) = Semicolon : Shell.lex t
lex s@(h : t)
	| isSpace h = Shell.lex t
	| is_ident_char h = Ident (takeWhile is_ident_char s) : Shell.lex (dropWhile is_ident_char s)

parse :: [Lexeme] -> ShellState -> IO (String, ShellState)
parse l state =
	case reverse l of
		(Semicolon : t) ->
			case (val_base v, leftover) of
				(VIO p, []) -> p state >>= \state' -> return ("success", state')
				(VError s, _) -> return ("error: " ++ s, state)
				(_, []) -> return ("error: wrong type", state)
				(_, t') -> error ("parse: pre-trailing lexemes " ++ show t' ++ " but no error")
			where
				(v, leftover) = parse' t
		_ -> return ("error: final lexeme is not Semicolon", state)

parse' :: [Lexeme] -> (Val, [Lexeme])
parse' [] = (Val id (VError "empty expression"), [])
parse' l@(LParen : _) = (Val id (VError "empty expression in parentheses"), l)
parse' (RParen : t) =
	case t' of
		(LParen : t') -> let (v', t'') = parse' t' in (val_f v' v, t'')
		_ -> (error_val "RParen without LParen", [])
	where
		(v, t') = parse' t
parse' (Semicolon : t) = (error_val "unexpected Semicolon", [])
parse' (Ident s : t) = let (v', t') = parse' t in (val_f v' (parse_ident s), t')

parse_ident :: String -> Val
parse_ident ('\'' : s) = base_val (VS s)
parse_ident "exit" = base_val (VIO (\s -> return (s {Shell.exit = True})))
parse_ident "render_rls" = Val f (VError "render_rls needs an argument")
	where
		f (Val _ (VScene scene)) = base_val (VIO (fork_render (RLS.muts scene 5e-2 1e-1)))
		f _ = error_val "render_rls: wrong argument type"
parse_ident "render_sps" = Val f (VError "render_sps needs an argument")
	where
		f (Val _ (VScene scene)) = base_val (VIO (fork_render (SPS.muts scene)))
		f _ = error_val "render_sps: wrong argument type"
parse_ident "scene" = Val f (VError "scene needs an argument")
	where
		f (Val _ (VS s)) = base_val (VScene (the_scenes s))
		f _ = error_val "scene: wrong argument type"
parse_ident "update" = base_val (VIO update_render)
parse_ident _ = error_val "unknown identifier"

fork_render :: MutationScheme StdGen d RGB (LP.T g, [Point]) -> ShellState -> IO ShellState
fork_render muts state =
	case render_info state of
		Nothing ->
			do	command_chan <- newChan
				response_chan <- newChan
				last_path_ref <- newIORef []
				forkIO (render muts (global_info state) command_chan response_chan)
				-- GLUT image_data <- mallocArray (3 * width * height)
				-- TODO: dynamic file name
				image_file <- openFile "image.ppm" WriteMode
				-- GLUT window <-
				-- GLUT	do	window_ref <- newIORef undefined
				-- GLUT		run_with_glut (global_info state) $ do
				-- GLUT			initialWindowSize $= size
				-- GLUT			createWindow "render window" >>= writeIORef window_ref
				-- GLUT			displayCallback $= display width height image_data last_path_ref
				-- GLUT		readIORef window_ref
				cpu_time <- getCPUTime
				real_time <- liftM clock_time_to_seconds getClockTime
				return
					(	state
					{	render_info = Just $
						RI
							command_chan
							response_chan
							last_path_ref
							image_file
							cpu_time
							real_time
					-- GLUT		image_data
					-- GLUT		window
					}
					)
		_ -> putStrLn "already rendering" >> return state

make_image :: (IArray a RGB) => a (Int, Int) RGB -> Ptr CUChar -> IO ()
make_image image img_data = pokeArray img_data (image_bytes False image)

update_render :: ShellState -> IO ShellState
update_render state =
	case render_info state of
		Nothing -> putStrLn "not rendering" >> return state
		Just ri ->
			do	writeChan (command_chan ri) (SendUpdate {rc_clear = False})
				response <- readChan (response_chan ri)
				putStrLn ("sample color: " ++ show (rr_last_color response))
				putStrLn ("sample path: " ++ show (rr_last_path response))
				putStrLn ("longest path length: " ++ show (rr_longest_path response))
				putStrLn ("mutations attempted: " ++ show (rr_n_mutations response))
				putStrLn ("mutations accetpd: " ++ show (rr_n_mutations_accepted response))
				cpu_time <- getCPUTime
				cpu_time_taken <- return (fromIntegral (cpu_time - start_cpu_time ri) * 1e-12)
				putStrLn ("render CPU time: " ++ show cpu_time_taken)
				clock_time <- liftM clock_time_to_seconds getClockTime
				putStrLn ("render real time: " ++ show (clock_time - start_clock_time ri))
				putStrLn ("mutations per second CPU: " ++ show (fromIntegral (rr_n_mutations response) / cpu_time_taken))
				write_PPM (image_file ri) (longImageToImage (rr_image response))
				-- GLUT make_image (rr_image response) (image_data ri)
				writeIORef (last_path_ref ri) (rr_last_path response)
				-- GLUT run_with_glut (global_info state) (postRedisplay (Just (window ri)))
				return state

shell :: GlobalInfo -> IO ()
shell global =
	do	shell' (SS global Nothing False)
		error "shell finished"
	where
		shell' state =
			do	putStr "bidding? "
				hFlush stdout
				command_lexemes <- liftM Shell.lex getLine
				(response, state') <- parse command_lexemes state
				putStrLn response
				if (Shell.exit state') then writeIORef (GlobalInfo.exit global) True else shell' state'
