module SubpathSample (T, muts) where

import Algebra
import Color
import Control.Monad.Trans
import qualified Data.Foldable as Fold
import Data.Sequence as Seq
import Geometry
import qualified LightProp as LP
import Math
import Maybe
import Metropolis
import Monad
import Part
import Rand
import Types
import Util

data (RandomGenD g) => Node g =
		Node
		{	node_loc :: Either End Point
		,	node_ebdrf :: EBDRF g
		,	node_ac :: Bool
		,	node_lp :: LP.T g
		}

data (RandomGenD g) => Edge g =
		Edge
		{	edge_dim :: LP.T g
		}

data T g =
		C
		{	nodes :: Seq (Node g)
		,	edges :: Seq (Edge g)
		}

muts :: (RandomGenD g) => Scene g -> MutationScheme g (T g) RGB (LP.T g, [Point])
muts scene = MutS eval (mutate scene) (initial scene)

max_path_length :: Int
max_path_length = 1000

eval :: (RandomGenD g) => T g -> (RGB, (LP.T g, [Point]))
eval path = (LP.color lp, (lp, catMaybes (Fold.toList (fmap (either (const Nothing) Just . node_loc) inner_nodes)))) where
	lp = Fold.foldl1 (@*) (fmap node_lp (nodes path)) @* Fold.foldl1 (@*) (fmap edge_dim (edges path))
	inner_nodes = Seq.drop 1 (Seq.take (pred (Seq.length (nodes path))) (nodes path))

mutate :: (RandomGenD g) => Scene g -> T g -> Rand g (T g, Double)
mutate scene path = retry (mutate' scene path)

mutate' :: (RandomGenD g) => Scene g -> T g -> MaybeT (Rand g) (T g, Double)
mutate' scene (C nodes edges) =
	do	(initial_index, final_index) <- lift rand_interval

		(nodes_before, rest) <- return $ Seq.splitAt initial_index nodes
		(old_nodes, nodes_after) <- return $ Seq.splitAt (succ final_index - initial_index) rest
		(edges_before, rest) <- return $ Seq.splitAt initial_index edges
		(_old_edges, edges_after) <- return $ Seq.splitAt (final_index - initial_index) rest

		initial_node <- return $ sfirst old_nodes
		initial_in_dir <- return $ case initial_index of
			0 -> error "SubpathSample: in_dir to initial node undefined"
			1 -> Left Source
			_ -> Right $ normalize
				(	from_right (node_loc initial_node)
				%-	from_right (node_loc (index nodes (pred initial_index)))
				)

		final_node <- return $ slast old_nodes
		final_in_dir <- return $ case Seq.length nodes - final_index of
			1 -> error "SubpathSample: in_dir to final node undefined"
			2 -> Left Source
			_ -> Right $ normalize
				(	from_right (node_loc final_node)
				%-	from_right (node_loc (index nodes (succ final_index)))
				)

		C new_nodes new_edges <- rand_path scene initial_in_dir initial_node final_in_dir final_node (max_path_length - Seq.length nodes_before - Seq.length nodes_after)
		return
			( C
				(nodes_before >< new_nodes >< nodes_after)
				(edges_before >< new_edges >< edges_after)
			,	subpath_prob old_nodes / subpath_prob new_nodes
			)
	where
		rand_interval =
			do	a <- rand_n (Seq.length nodes)
				b <- rand_n (Seq.length nodes - 1)
				return $ if a <= b then (a, succ b) else (b, a)
		sfirst s = let x :< _ = viewl s in x
		slast s = let _ :> x = viewr s in x

initial :: (RandomGenD g) => Scene g -> Rand g (T g)
initial scene = retry path where
	path =
		rand_path
			scene
			(error "SubpathSample: initial: initial in_dir undefined")
			first_node
			(error "SubpathSample: initial: final in_dir undefined")
			last_node
			max_path_length
	first_node =
		Node
			(Left Source)
			source_ebdrf
			True
			(error "SubpathSample: initial: first lp undefined")
	last_node =
		Node
			(Left Sink)
			sink_ebdrf
			True
			(error "SubpathSample: initial: last lp undefined")

source_ebdrf, sink_ebdrf :: (RandomGenD g) => EBDRF g
source_ebdrf = const (point_mass (Left Source) idvec)
sink_ebdrf = const (point_mass (Left Sink) idvec)

rand_path :: (RandomGenD g) =>
		Scene g
	->	Either End Vec
	->	Node g
	->	Either End Vec
	->	Node g
	->	Int
	->	MaybeT (Rand g) (T g)
rand_path scene initial_in_dir initial_node final_in_dir final_node max_length =
	do	(fwd_nodes, fwd_edges, fwd_final_indir, fwd_final_node) <-
			rand_half_path scene initial_node initial_in_dir empty empty 0 (max_length - 1)
		(bwd_nodes, bwd_edges, bwd_final_indir, bwd_final_node) <-
			rand_half_path scene final_node final_in_dir empty empty 0 (max_length - Seq.length fwd_nodes)
		fwd_dir_m <- return $ node_ebdrf fwd_final_node fwd_final_indir
		bwd_dir_m <- return $ node_ebdrf bwd_final_node bwd_final_indir
		guard (ac_p_ fwd_dir_m > eps && ac_p_ bwd_dir_m > eps)
		(_, dim_lp, join_dir, join_undir) <- join_or_block scene (node_loc fwd_final_node) (node_loc bwd_final_node)
		fwd_final_node' <- return $
			fwd_final_node
			{	node_ac = True
			,	node_lp =
					ac_f_ fwd_dir_m join_dir
				.*	(ac_p_ fwd_dir_m * ac_rf_ fwd_dir_m join_dir)
			}
		bwd_final_node' <- return $
			bwd_final_node
			{	node_ac = True
			,	node_lp =
					ac_f_ bwd_dir_m join_undir
				.*	(ac_p_ bwd_dir_m * ac_rf_ bwd_dir_m join_undir)
			}
		guard $ normSqr (LP.color (node_lp fwd_final_node')) > eps && normSqr (LP.color (node_lp bwd_final_node')) > eps
		return $ C
			(fwd_nodes >< fwd_final_node' <| bwd_final_node' <| Seq.reverse bwd_nodes)
			(fwd_edges >< (Edge dim_lp) <| Seq.reverse bwd_edges)

rand_half_path :: (RandomGenD g) =>
		Scene g
	->	Node g
	->	Either End Vec
	->	Seq (Node g)
	->	Seq (Edge g)
	->	Int
	->	Int
	->	MaybeT (Rand g) (Seq (Node g), Seq (Edge g), Either End Vec, Node g)
rand_half_path scene start in_dir node_accum edge_accum n_steps_taken max_length =
	do	last_step <- lift $ bias_bool p_last
		if last_step
			then return (node_accum, edge_accum, in_dir, start {node_ac = True})
			else
			do	ac <- lift $ bias_bool (ac_p_ dir_m)
				(lp, dir) <- lift $ (if ac then ac_r_ else ms_r_) dir_m
				guard $ normSqr (LP.color lp) > eps

				ray <- case node_loc start of
					(Left s) ->
						do	-- make sure the direction is the same as s
							guard $ either (s ==) (const False) dir
							return $ Left s
					(Right source_p) ->
						do	-- Make sure dir is a vector (Right), not an End (Left).
							-- (A rand_half_path can start at an End, but can't end at one.)
							dir_vec <- MaybeT $ return $ either (const Nothing) Just dir
							return $ Right (source_p, dir_vec)

				(hit_r, dim) <- return $ trace (part_ scene) ray
				hit <- hit_r
				new_node <- return $
					Node
						(Right (hit_loc hit))
						(hit_ebdrf hit)

						-- ac and lp are set when the next segment of the path is generated
						(error "SubpathSample: node ac undefined")
						(error "SubpathSample: node lp undefined")
				new_edge <- return $ Edge (if is_right (node_loc start) then dim (hit_dist hit) else idvec)
				rand_half_path
					scene
					new_node
					dir
					(node_accum |> start {node_ac = ac, node_lp = lp})
					(edge_accum |> new_edge)
					(succ n_steps_taken)
					(max_length - 1)
	where
		p_last = if max_length == 1 then 1 else 1 / fromIntegral (n_steps_taken + 2)
		dir_m = node_ebdrf start in_dir

subpath_prob :: (RandomGenD g) => Seq (Node g) -> Double
subpath_prob nodes = sum [prob_at i | i <- [0 .. n - 2]] where
	n = Seq.length nodes
	prob_at i =
		if node_ac fwd_final_node && node_ac bwd_final_node
			then
				length_prob
			*	ac_rf_ fwd_dir_m join_dir
			*	ac_rf_ bwd_dir_m join_undir
			*	join_distSqr
			else 0
		where
		length_prob = 1 / fromIntegral ((i + 1) * (i + 2) * (n - i) * (n - i + 1))
		fwd_final_node = index nodes i
		bwd_final_node = index nodes (succ i)
		(join_distSqr, join_dir, join_undir) = join_info (node_loc fwd_final_node) (node_loc bwd_final_node)
		fwd_dir_m = node_ebdrf fwd_final_node join_dir
		bwd_dir_m = node_ebdrf bwd_final_node join_undir

join_info :: Either End Vec -> Either End Vec -> (Double, Either End Vec, Either End Vec)
join_info (Left _) (Left _) = error "join_dir: two Ends joined"
join_info s@(Left _) _ = (1, s, s)
join_info _ s@(Left _) = (1, s, s)
join_info (Right u) (Right v) = (normSqr diff, Right dir, Right (dir .* (- 1))) where
	diff = v %- u
	dir = normalize diff

-- TODO: this has a lot in common with join_info
join_or_block :: (RandomGenD g) => Scene g -> Either End Vec -> Either End Vec -> MaybeT (Rand g) (Double, LP.T g, Either End Vec, Either End Vec)
join_or_block _ (Left _) (Left _) = error "join_dir: two Ends joined"
join_or_block _ s@(Left _) _ = return (1, idvec, s, s)
join_or_block _ _ s@(Left _) = return (1, idvec, s, s)
join_or_block scene (Right u) (Right v) =
	do	hit_m <- lift $ run_maybe_t hit_r
		guard $ fromMaybe True $ liftM ((dist <) . (+ eps) . hit_dist) hit_m
		return (distSqr, dim dist, Right dir, Right (dir .* (- 1)))
	where
	diff = v %- u
	distSqr = normSqr diff
	dist = sqrt distSqr
	dir = diff ./ dist
	(hit_r, dim) = trace (part_ scene) (Right (u, dir))
