module Main where

import CUtil
import Color
import Control.Concurrent
import Control.Monad.Error
import Control.Monad.Identity
import qualified Data.ByteString as BS
import Data.IORef
import qualified Data.String.UTF8 as UTF8
import Directory
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable
import Geometry
import GlobalInfo
import IO
import JobDescription
import qualified LightProp as LP
import Metropolis
import Monad
import qualified RListSample as RLS
import Rand
import RasterUtil
import Render
import qualified SubpathSample as SPS
import System
import System.Posix.Files
import System.Posix.IO
import System.Posix.Types
import qualified Text.SExpression as SE
import Text.SExpression.Format
import Util

data Stats =
    Stats
    { n_mutations, n_mutations_accepted :: Integer
    } deriving (Read, Show)

main :: IO ()
main =
    do jobDescStr <- readUTF8File "job_desc"
       jobDesc <- case runErrorT (readDesc jobDescStr)  of
           Identity (Right x) -> return x
           Identity (Left err) -> do putStrLn ("error parsing the job description: " ++ err)
                                     exitFailure
       when ("0" /= renderVersion jobDesc) $ fail "bad version"
       exitRef <- newIORef False
       commandChan <- newChan
       responseChan <- newChan
       lockFd <- openFd "lock" ReadOnly Nothing defaultFileFlags
       render' <- buildMutationScheme jobDesc (\m -> render m (GI {exit = exitRef}) commandChan responseChan)
       forkIO render'
       jobLoop jobDesc commandChan responseChan lockFd

jobLoop :: JobDescription -> Chan RenderCommand -> Chan RenderResponse -> Fd -> IO ()
jobLoop jobDesc commandChan responseChan lockFd =
    do threadDelay 10000000
       putStrLn "dumping latest points..."
       putStrLn "  acquiring lock..."
       lockFile lockFd
       putStrLn "  waiting for render thread..."
       writeChan commandChan (SendUpdate {rc_clear = True})
       renderResponse <- readChan responseChan
       putStrLn "  updating the files..."
       do rasterExists <- doesFileExist "raster"
          statsExists <- doesFileExist "stats"
          unless (rasterExists && statsExists) $ fail "jobLoop: raster or stats file doesn't exist"
       renameFile "raster" "raster.bak"
       renameFile "stats" "stats.bak"

       -- Update raster file.
       let rasterSize = imageWidth jobDesc * imageHeight jobDesc * max (sizeOf (undefined :: LongRGB)) (alignment (undefined :: LongRGB))
       putStrLn "    reading the old raster..."
       raster <-
           do fd <- openFd "raster.bak" ReadOnly Nothing defaultFileFlags
              ptr <- mallocBytes rasterSize
              readToPtr fd ptr (fromIntegral rasterSize)
              closeFd fd
              ret <- (readStorableArray (castPtr ptr) ((0, 0), (imageWidth jobDesc - 1, imageHeight jobDesc - 1)))
              free ptr
              return ret
       addRasterTo raster (rr_image renderResponse)
       putStrLn "    writing the new raster..."
       do fd <- createFile "raster" ownerReadMode
          ptr <- mallocBytes rasterSize
          writeStorableArray (castPtr ptr) raster
          writeFromPtr fd ptr (fromIntegral rasterSize)
          closeFd fd
          free ptr

       -- Update stats file.
       putStrLn "    updating the stats file..."
       oldStats <- liftM read $ readFile "stats.bak"
       let newStats = oldStats
             { n_mutations = n_mutations oldStats + rr_n_mutations renderResponse
             , n_mutations_accepted = n_mutations_accepted oldStats + rr_n_mutations_accepted renderResponse
             }
       writeFile "stats" (show newStats)

       putStrLn "  releasing lock..."
       unlockFile lockFd

       putStrLn "done."

       jobLoop jobDesc commandChan responseChan lockFd

buildMutationScheme :: (RandomGenD g, Monad m) => JobDescription -> (forall d g'. MutationScheme g d RGB (LP.T g', [Point]) -> a) -> m a
buildMutationScheme jd f =
    case sampleMethod jd of
        "rlist" -> return (f (RLS.muts (scene jd) 5e-2 1e-1))
        "sps" -> return (f (SPS.muts (scene jd)))
        s -> fail ("buildMutationScheme: unknown scheme name: " ++ s)
